import java.io.*; 
import java.util.*;
import java.nio.*;
import java.nio.channels.*;
import javax.imageio.stream.*;
import javax.swing.*;
import java.awt.*;
import java.util.regex.*;

import cj.appkit.*;
import cj.math.jama.*;
import cj.fileformats.*;
import gnu.getopt.*;

class PARREC
{
	private ShortBuffer shBuf = null;
	private String rec_filename = null;
	private String par_filename = null;

	private int parVersion = 0;
	private boolean do_scaling = true;

	// Image information
	private int nrows = 0;
	private int ncols = 0;
	private int nslices = 0;
	private int nchannels = 0;

	//  Scaling factors
	private double[] scaling_ri = new double[1];
	private double[] scaling_rs = new double[1];
	private double[] scaling_ss = new double[1];

	// b values
	private double[] b_value = new double[1];
	
	public PARREC(String filename, boolean do_scaling)
	{
		if( filename.indexOf(".REC")  > filename.indexOf(".rec"))
		{
			rec_filename = filename;
            par_filename  = rec_filename.substring(0,rec_filename.lastIndexOf(".REC"))+".PAR";
            //par_filename = rec_filename.replaceAll("REC", "PAR");
		}
		else if( filename.indexOf(".rec") > 0 )
		{
			rec_filename = filename;
            par_filename  = rec_filename.substring(0,rec_filename.lastIndexOf(".rec"))+".par";
            //par_filename = rec_filename.replaceAll("rec", "par");
		}
		else
		{
            throw new RuntimeException("PARREC: Unable to identify rec file extension (rec or REC).");
        }

		this.do_scaling = do_scaling;

		parsePAR();
		openREC();

		nchannels = b_value.length;
	}

	private void openREC()
	{
		//--------------------------------------------------------------------
		//
		// Open the file.
		//
		//--------------------------------------------------------------------
		try {
			FileChannel roChannel = new 
				RandomAccessFile(rec_filename, "r").getChannel();
			shBuf = roChannel.map(FileChannel.MapMode.READ_ONLY, 0,
				  (int)roChannel.size()).order(ByteOrder.LITTLE_ENDIAN).asShortBuffer();

		}
		catch(Exception e) { System.out.println(e); System.exit(-1); }
	}

	//  This will retrieve the data from the file
	//  and will include all the scaling etc.
	final public double[] getData(int row, int col, int slice)
	{
		double[] all_data = new double[nchannels+1]; // bug fix - rest of code is based on 1 access via skip map
		Arrays.fill(all_data, 0.0);

		//  Compute the gradient information
		for(int ii=0; ii<(nchannels+1); ii++) 
		{ 
			int scalar_index =  ii*nslices + slice;

			double tt = (double)shBuf.get( ii*nrows*ncols*nslices + 
											slice * nrows*ncols + 
											row * ncols +
											col );
			if( do_scaling ) 
			{
					tt = (tt * scaling_rs[scalar_index] + scaling_ri[scalar_index]) /
						(scaling_rs[scalar_index]*scaling_ss[scalar_index] );
			}
			
			all_data[ii] = tt;
		}

		return all_data;
	}

	final public int getNRows() { return nrows; }
	final public int getNCols() { return ncols; }
	final public int getNSlices() { return nslices; }
	final public double[] getBValues() { return b_value; }

	//
	//  Get the number of rows, columns and slices from the PAR file.
	//
	private void parsePAR()
	{
		System.out.println("Going to parse " + par_filename);

		Pattern pattern_version = 
			Pattern.compile(".*\\s+V([34]).*");

		ImageInputStream ldis = AppKit.openRead(par_filename);

		String line = new String();

		boolean done = false;
		do
		{
			try {
				line = ldis.readLine();	

				if( line == null ) { continue; }

				Matcher matcher_version = pattern_version.matcher(line);
				if( matcher_version.matches() && parVersion == 0 )
				{ 
					parVersion = Integer.parseInt(matcher_version.group(1)); 
					System.out.println("PAR file is version " + parVersion );
				}

				if( parVersion == 3 )
					{ parseVersion3(ldis, line); break;}
				else if ( parVersion == 4 )
					{ parseVersion4(ldis, line); break;}

			}
			catch(EOFException e) { done = true; }
			catch(IOException e) { System.out.println(e); done = true; }
		}
		while( !done && line != null );
	}

	final private void parseVersion3(ImageInputStream ldis, String line)
	{
		boolean done = false;

		Pattern pattern_hw = Pattern.compile(".+Recon resolution.*:\\s*([0-9]*)\\s*([0-9]*).*");
		Pattern pattern_slices = Pattern.compile(".+Max. number of slices/locations.*:\\s*([0-9]*).*");
		Pattern numbers = Pattern.compile("(\\s+[\\d\\.\\-e]+)+\\s*");
		Pattern number2 = Pattern.compile("(\\s*([\\d\\-\\.e]+)){7,}\\s*");
		Pattern spaces = Pattern.compile("\\s+");

		ArrayList<Double> rs = new ArrayList<Double>();
		ArrayList<Double> ri = new ArrayList<Double>();
		ArrayList<Double> ss = new ArrayList<Double>();
		ArrayList<Double> bb = new ArrayList<Double>(); // b value

		do
		{
			try {
				line = ldis.readLine();	

				if( line == null )
                {
                    continue;
                }
                //System.out.println(line);

                Matcher matcher_hw = pattern_hw.matcher(line);
				if( matcher_hw.matches() )
				{ 
					nrows = Integer.parseInt(matcher_hw.group(1));
					ncols = Integer.parseInt(matcher_hw.group(2));
				}

				Matcher matcher = number2.matcher(line);

				if( matcher.matches() )
				{ 
					String[] bob = spaces.split(line.trim());

					ri.add( Double.parseDouble(bob[7]) );
					rs.add( Double.parseDouble(bob[8]) );
					ss.add( Double.parseDouble(bob[9]) );

					//  Remove the 0's here.
					if( Double.parseDouble( bob[27] ) != 0.0 )
					{
						bb.add( Double.parseDouble( bob[27] ) );
					}
				}

				Matcher matcher_slices = pattern_slices.matcher(line);
				if( matcher_slices.matches() )
				{ 
					nslices = Integer.parseInt(matcher_slices.group(1)); 
				}
			}
			catch(EOFException e) { done = true; }
			catch(IOException e) { System.out.println(e); done = true; }
		}
		while( !done && line != null );

		//  Copy the scaling factors over to the variables.
		scaling_ri = new double[ ri.size() ];
		scaling_rs = new double[ ri.size() ];
		scaling_ss = new double[ ri.size() ];
		b_value = new double[ bb.size()/nslices ];

		for(int ii=0; ii< ri.size(); ii++) 
		{
			scaling_ri[ii] = ri.get(ii);
			scaling_rs[ii] = rs.get(ii);
			scaling_ss[ii] = ss.get(ii);

		}

		/// Calculate the mean b value so that it 
		//  can be replaced into the array of b -values
		//  where there is the "-1".  Sometimes found
		//  in old 1.5T data.
		double meanvalue = 0.0; int nn = 0;
		for(int ii=0; ii< bb.size(); ii++) 
		{
			if( bb.get(ii) > 0.0 ){ meanvalue += bb.get(ii); nn++; }
		}
		meanvalue /= (double)nn;

		for(int ii=0, jj=0; ii< bb.size(); ii+=nslices, jj++) 
		{
			if( bb.get(ii) < -0.1 )
			{
				b_value[jj] = meanvalue;
			}
			else
				b_value[jj] = bb.get(ii);
		}
	}

	final private void parseVersion4(ImageInputStream ldis,String line)
	{ 
		boolean done = false;

		Pattern pattern_slices = 
		Pattern.compile(".+Max. number of slices/locations.*:\\s*([0-9]*).*");
		Pattern numbers = Pattern.compile("(\\s+[\\d\\.\\-e]+)+\\s*");
		Pattern number2 = Pattern.compile("(\\s*([\\d\\-\\.e]+)){7,}\\s*");
		Pattern spaces = Pattern.compile("\\s+");

		ArrayList<Double> rs = new ArrayList<Double>();
		ArrayList<Double> ri = new ArrayList<Double>();
		ArrayList<Double> ss = new ArrayList<Double>();
		ArrayList<Double> bb = new ArrayList<Double>(); // b value

		do
		{
			try {
				line = ldis.readLine();	

				if( line == null ) { continue; }

				Matcher matcher_slices = pattern_slices.matcher(line);
				if( matcher_slices.matches() )
				{ 
					nslices = Integer.parseInt(matcher_slices.group(1)); 
				}

				Matcher matcher = number2.matcher(line);

				if( matcher.matches() )
				{ 
					String[] bob = spaces.split(line.trim());

					ri.add( Double.parseDouble(bob[11]) );
					rs.add( Double.parseDouble(bob[12]) );
					ss.add( Double.parseDouble(bob[13]) );

					//  Remove the 0's here.
					if( Double.parseDouble( bob[33] ) > 0.0 )
						bb.add( Double.parseDouble( bob[33] ) );
				}

				if( nrows == 0 && matcher.matches() )
				{ 
					String[] bob = spaces.split(line.trim());

					nrows = Integer.parseInt(bob[9]); 
					ncols = Integer.parseInt(bob[10]); 
				}
			}
			catch(EOFException e) { done = true; }
			catch(IOException e) { System.out.println(e); done = true; }
		}
		while( !done && line != null );

		//  Copy the scaling factors over to the variables.
		scaling_ri = new double[ ri.size() ];
		scaling_rs = new double[ ri.size() ];
		scaling_ss = new double[ ri.size() ];
		b_value = new double[ bb.size() / nslices ];

		for(int ii=0; ii< ri.size(); ii++) 
		{
			scaling_ri[ii] = ri.get(ii);
			scaling_rs[ii] = rs.get(ii);
			scaling_ss[ii] = ss.get(ii);
		}

		for(int ii=0, jj=0; ii< bb.size(); ii+=nslices, jj++) 
		{
			b_value[jj] = bb.get(ii);
		}
	}

	public double getPercentileThreshold( double percentile )
	{
		double[] blah = new double[nrows*ncols*nslices];
		int ii=0;

		for(int row=0; row<nrows; row++) 
		{ 
			for(int col=0; col<ncols; col++) 
			{ 
				for(int slice=0; slice<nslices; slice++) 
				{ 
					blah[ii] =  (double)shBuf.get( slice * nrows*ncols + 
												  row * ncols +
												  col ); 
					if( do_scaling ) 
					{ 
						blah[ii] = (blah[ii] * scaling_rs[slice] + scaling_ri[slice]) / 
							 (scaling_rs[slice]*scaling_ss[slice]);
					}

					ii++;
				}
			}
		}

		return getPercentile( blah, percentile );	
	}

	private double getPercentile(double[] array, double percentile)
	{
		if (percentile == 0)
			return array[0];

		Arrays.sort(array);

		int idx = (int) ((double) array.length * ((double) percentile / 100.0));
		
		return array[idx];
	}
}

