package apps.decay;

import java.lang.*;
import java.util.*;
import gnu.getopt.*;
import cj.mri.hennig.*;

class DecayApp { 

	public static void usage()
	{
		System.out.println("decay [options] [num echoes]");
		System.out.println("Options:");
        System.out.println("\t[-h|--help] - This help");
        System.out.println("\t[-e|--te] - Echo time (in ms, def: 10)");
        System.out.println("\t[-r|--tr] - Repition time (in ms, def: inf)");
        System.out.println("\t[-l|--t1] - T1 time (in ms, def: 1000)");
        System.out.println("\t[-s|--t2] - T2 time (in ms, def: 80)");
        System.out.println("\t[-a|--excite-flip] - Excitation pulse (in degrees, def: 90)");
        System.out.println("\t[-b|--refocus-flip] - Refocusing pulse (in degrees, def: 180)");
        System.out.println("\t[-c|--crushers] - Effectiveness of crushers (0..1, def: 0)");
        System.out.println("\t[-p|--pd] - Proton density weighting (def: 1.0)");
        System.out.println("\t[-f|--first-echo] - First echo weighting (def: 1.0)");
        System.out.println("\t[-t|--timing] - Print out timing");

        System.out.println("");
		System.out.println("Calculate decay curve given refocusing pulse of alpha");
        System.out.println("\tNotes: ");
        System.out.println("\t    - '-p' and '-f' are mutually exclusive.");
	}

	public static void main(String args[]) {
        short verbose = 0;
 		double te = 10.0;
		double tr = Double.POSITIVE_INFINITY;
		double t1 = 1000.0;
		double t2 = 80.0;
		double exciteFlip = 90.0;
		double refocusFlip = 180.0;
		double crushers = 0.0;
		double pd = 1.0;
		double first_echo = 1.0;
		int num_echoes = 10;
		boolean doTiming = false;
		long start = 0, end = 0;
 
        //-------------------------------------------------------
        //
        //  Check the arguements to the program.
        //
        LongOpt[] longopts = new LongOpt[11];
        longopts[0] = new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h');
        longopts[1] = new LongOpt("te", LongOpt.REQUIRED_ARGUMENT, null, 'e');
        longopts[2] = new LongOpt("tr", LongOpt.REQUIRED_ARGUMENT, null, 'r');
        longopts[3] = new LongOpt("t1", LongOpt.REQUIRED_ARGUMENT, null, 'l');
        longopts[4] = new LongOpt("t2", LongOpt.REQUIRED_ARGUMENT, null, 's');
        longopts[5] = new LongOpt("excite-flip", LongOpt.REQUIRED_ARGUMENT, null, 'a');
        longopts[6] = new LongOpt("refocus-flip", LongOpt.REQUIRED_ARGUMENT, null, 'b');
        longopts[7] = new LongOpt("crushers", LongOpt.REQUIRED_ARGUMENT, null, 'c');
        longopts[8] = new LongOpt("pd", LongOpt.REQUIRED_ARGUMENT, null, 'p');
        longopts[9] = new LongOpt("first-echo", LongOpt.REQUIRED_ARGUMENT, null, 'f');
        longopts[10] = new LongOpt("timing", LongOpt.NO_ARGUMENT, null, 't');

        Getopt g = new Getopt("decay", args, "hvte:r:s:l:b:a:c:p:f:", longopts);
        int c;
        while((c=g.getopt()) != -1 ) {
            switch(c) {
                case 'h':
					usage();
                    System.exit(0);
                break;
                case 'v':
                    verbose++;
                break;
                case 't':
                    doTiming = true;
                break;
                case 'e':
                    te = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'r':
                    tr = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'l':
                    t1 = (new Double(g.getOptarg())).doubleValue();
                break;
                case 's':
                    t2 = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'p':
                    pd = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'f':
                    first_echo = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'a':
                    exciteFlip = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'b':
                    refocusFlip = (new Double(g.getOptarg())).doubleValue();
                break;
                case 'c':
                    crushers = (new Double(g.getOptarg())).doubleValue();
                break;
            }
        }

		if( args.length < 1 ) 
		{
			usage();
			System.exit(0);
		}

        if( g.getOptind() < args.length ) {
            num_echoes = (new Integer(args[g.getOptind()])).intValue();
        }

		if( verbose > 0 )
		{
			System.out.println("Decay curve with: ");
			System.out.println("\tTE = " + te + " ms");
			System.out.println("\tTR = " + tr + " ms");
			System.out.println("\tT1 = " + t1 + " ms");
			System.out.println("\tT2 = " + t2 + " ms");
			System.out.println("\tPD = " + pd);  
			System.out.println("\tFirst echo = " + first_echo);  
			System.out.println("\texcitation pulse flip angle = " + exciteFlip + " degrees");
			System.out.println("\trefocusing pulse flip angle = " + refocusFlip + " degrees");
			System.out.println("\tcrushers = " + crushers);
		}

		Decay decay = new Decay(num_echoes);
		decay.setTE( te );
		decay.setTR( tr );
		decay.setT1( t1 );
		decay.setT2( t2 );
		decay.setExciteFlip( exciteFlip );
		decay.setRefocusFlip( refocusFlip );
		decay.setCrushers( crushers );
		decay.setPD( pd );
		decay.setFirstEcho( first_echo );
		decay.setVerbose( verbose );

		if( doTiming )
		{
			start = System.currentTimeMillis();
		}

		double [] signal = decay.calculateEchoes( );

		if( doTiming )
		{
			end = System.currentTimeMillis();
		}

		for(int ii=0; ii<signal.length; ii++) 
		{
			System.out.println(signal[ii]);
		}

		if( doTiming )
		{
			System.out.println("This took "+((end-start)/1000.0) + " seconds");
		}
	}
}
