package apps.rephase;

import gnu.getopt.*;
import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.volume.process.Rephase;
import cj.appkit.FilterAppKit;

final public class RephaseApp {
   public static void main(String argv[]) {
      RephaseAppKit myApp = new RephaseAppKit(); myApp.main(argv);
   }
}

final class RephaseAppKit extends FilterAppKit {

	private int uniform_blur_window = 17;
	private int blur_iterations = 3;
	private double gaussian_sigma = 3.0;
	private int blur_method = Rephase.BLUR_UNIFORM;
	private boolean remove_imaginary = true;
	private Rephase re = null;
    private String comment;
	private String output_suffix = "_rephased";

	public String getAppName() { return "mrrephase"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

	public void help()
	{
		super.help();

		System.out.println("Notes: ");
		System.out.println("   --blur-type can be:");
		System.out.println("      uniform <window size> (e.g. --blur-type uniform 3)");
		System.out.println("      gaussian <sigma>      (e.g. --blur-type gaussian 4.5)");
	}

    protected void moreOpts() { 
        addOpt_reqArg("blur-iterations", 'b', 
			"Number of blurring iterations (def: 3).");
        addOpt_reqArg("blur-type", 't', 2, "Blur type (def: uniform 17 ).");
        addOpt_noArg("keep-imaginary", 'k', 
			"Keep imaginary images (def: remove).");
        addOpt_reqArg("output-suffix", 800, "Output suffix (def: '_rephased').");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case 800:
				output_suffix = getStringArg("suffix");
				break;
			case 'k':
				remove_imaginary = false;
				break;
			case 't':
				/*
				 *  Get the type of blurring.
				 */
				String blur_type = getStringArg("blur type");

				if( blur_type.toLowerCase().compareTo("uniform") == 0 )
				{
					blur_method = Rephase.BLUR_UNIFORM;
					uniform_blur_window = getIntArg("uniform window");
				}
				else if( blur_type.toLowerCase().compareTo("gaussian") == 0 )
				{
					blur_method = Rephase.BLUR_GAUSSIAN;
					gaussian_sigma = getDoubleArg("Gaussian sigma");
				}
				else
				{
					System.out.println(getAppName() + ": Unknown blur type " + blur_type);
					System.exit(-1);
				}

				break;
			case 'b':
				blur_iterations = getIntArg("number of blur iterations");
				break;
			default:
				help();
                System.exit(-1);
		}

		return true;
    }

	final public void initParams(MCVolume v)
	{
		re = new Rephase();
		re.setBlurIterations( blur_iterations );
		re.setBlurMethod( blur_method );
	}

    final public MCVolume filter(MCVolume v)
	{
		re.rephase(v);

		return v;
	}

	public void preSave( MRIFile b )
	{
		super.preSave(b);

		if( b.getFileType() == MRIFile.FILETYPE_BFF )
		{ 
			if( remove_imaginary )
			{
				int nchannels = b.getMCVolume().getNChannels();
				for(int ii=1; ii<=nchannels/2; ii++)
				{
					if( verbose > 1 )
					{
						System.out.println("Removing volume " + ii);
					}

					((BFF)b).removeVolume(ii); 
				}
			}
		}

	}
}
