package apps.sinormalise;

import gnu.getopt.*;
import cj.fileformats.*;
import cj.volume.*;
import cj.volume.process.*;
import cj.appkit.FilterAppKit;

final public class SINormaliseApp {
   public static void main(String argv[]) {
      new SINormaliseAppKit().main(argv);
   }
}

final class SINormaliseAppKit extends FilterAppKit {

	boolean train_f = false;
	boolean correct_f = false;
	double standardized_mean = -1.0;
	double percent = 99.8;
	double max = 4095.0;
	String correction_method = "gamma";

	SINormalise[] uc = null;

	// Needed as a hack to do the two pass thing in filter()
	boolean fromPost = false;
	private String comment = "";

	public String getAppName() { return "mrsinorm"; }
	public String getDialogName() { return "Process image"; }
	public String getOutputSuffix() { return "U"; }
	public String getComment() { return comment; }

    public void help() 
	{
		super.help();
		System.out.println("Valid option combinations are:");
		System.out.println("    -t -c -> Train then correct.");
		System.out.println("    -s <num> -c -> Correct based on standardized mean passed in.");
		System.out.println("    -t -> Train and write out the standardized mean.");
    }

    protected void moreOpts() 
	{ 
		addOpt_noArg("train", 't', "Train (def: no).");
		addOpt_noArg("correct", 'c', "Correct (def: no).");
		addOpt_reqArg("percentile", 'p', "Set upper percentile (def. 99.8)");
		addOpt_reqArg("maximum", 'M', "Set maximum intensity (def. 4095)");
		addOpt_reqArg("standardized-mean", 's', 
		            "Set the standardized mean (def: calculated in training step).");
		addOpt_reqArg("correct-method", 'm', 
		              "Set the correction method (def: gamma).");
    }

    protected boolean parseArg(int c)
    {
    	switch(c)
		{
			case 't':
				train_f = true;	
			break;
			case 'c':
				correct_f = true;
			break;
            case 'p':
				percent = getDoubleArg("percentile");
			break;
            case 'M':
				max = getDoubleArg("maximum");
			break;
			case 's':
				standardized_mean = getDoubleArg("standardized mean");
			break;
			case 'm':
				correction_method = getStringArg("correct method");
			break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams(MCVolume v)
	{
		/*
		 *  Do some sanity checking.
		 */
		if( train_f && standardized_mean > 0.0 )
		{
			System.out.print("mrsinorm: You specified training and a standardized ");
			System.out.println("mean, the standardized mean is going to be ignored.");
		}

		if( correct_f && !train_f && standardized_mean < 0.0 )
		{
			System.out.println("sirnom: You must either train or pass in a standardized mean.");
			help();
			System.exit(-1);
		}

		if( uc == null ) 
		{
			if( verbose > 0 )
			{
				System.out.println("sinorm: Initializing parameters.");
				System.out.println("Upper limit at "+percent+"th percentile.");
			}

			/*
			 *  Initialize the SINormalize objects.
			 */
			uc = new SINormalise[v.getNChannels()];
			for(int ii=0; ii<v.getNChannels(); ii++) 
			{
				uc[ii] = new SINormalise();

				if( correction_method.equals("gamma") )
				{
					uc[ii].setMethod( SINormalise.METHOD_GAMMA );
				}
				else if( correction_method.equals("udupa") )
				{
					uc[ii].setMethod( SINormalise.METHOD_UDUPA );
				}
				else if( correction_method.equals("simple") )
				{
					uc[ii].setMethod( SINormalise.METHOD_SIMPLE );
				}
				else
				{
					System.out.println("IntensityCorrect: Unknown correction method "
									+ correction_method);
					System.exit(-1);
				}

				if( standardized_mean > 0.0 )
				{
					uc[ii].setStandardizedMean( standardized_mean );
				}
				if (percent > 0.0)
				{
					uc[ii].setPC2(percent);
				}
				if (max > 0.0)
				{
					uc[ii].setS2(max);
				}
			}
		}
	}

    final public void printArgs()
	{
	}

	public void pre()
	{
		if ( train_f && !fromPost ) filterMode = READ_ONLY;
	}
	// 
	//  Overridden post() method that calls run again, this time
	//  to do the correction and write out the images.
	//
	public void post()
	{
		if( correct_f && !fromPost ) 
		{
			fromPost = true;
			filterMode = READ_WRITE;
			run();  // Major hack
		}
	}

	final public MCVolume filter(MCVolume vol) 
	{
		/*
		 *  Set the verboseness for each correction object.
		 */
		for(int channel=0; channel<vol.getNChannels(); channel++)
		{
			uc[channel].setVerbose( verbose );
		}
		
		/*
		 *  TRAINING
		 */
		if( train_f && !fromPost ) 
		{
			if( verbose > 1 ) { System.out.println("sinorm: Training on "+filename+"..."); }
			for(int channel=0; channel<vol.getNChannels(); channel++)
			{
				uc[channel].train(vol.getVolume(channel));
			}

			if( !correct_f ) 
			{
				//if( verbose > 1 ) { System.out.println("sinorm: Standardized mean: "); }
				for(int channel=0; channel<vol.getNChannels(); channel++)
				{
					System.out.println("\tChannel " + (channel+1) + 
						": " + uc[channel].getStandardizedMean());
				}
			}
		}

		/*
		 *  Correct - only done after the post() is called
		 *  and if the correcting is necessary
		 */
		if( fromPost && correct_f ) 
		{
			if( verbose > 1 ) { System.out.println("sinorm: Correcting "+filename+"..."); }
			for(int channel=0; channel<vol.getNChannels(); channel++)
			{
				vol.setVolume(uc[channel].correct(vol.getVolume(channel)), channel);
				//comment = "udupa: standardized mean is " + uc[channel].getStandardizedMean(); 
			}
		}

		return vol;
	}
}
