package apps.t2map;

import cj.appkit.*;
import cj.volume.*;
import cj.fileformats.*;
import cj.mri.relax.t2map.*;
import java.io.File;
import java.util.*;
import java.util.*;
import gnu.getopt.*;

/**
 *  This is the T2 application.
 */
public class T2MapApp {
	public static void main(String argv[]) {
		T2MapAppFilter myApp = new T2MapAppFilter(); 
		myApp.main(argv);
	}
}


/** T2 Mapping class that calculates the T2 amplitudes and times
 *  for a given MCVolume.
 */
class T2MapAppFilter extends FilterAppKit
{ 
	private String output_directory = "./";
	private Vector bff_header = null;

	private String fit_type = "nnls";
	private double[] fit_params = null;

	private double threshold = 3.0; 
	private boolean normalizeFlag = true;
	private int rowStart = -1;
	private int rowEnd = -1;
	private int colStart = -1;
	private int colEnd = -1;
	private int sliceStart = -1;
	private int sliceEnd = -1;

	private String ne_method = "row";

	private T2Map mapper = null;

	private String comment = "T2Map ";

	/** Do we need to estimate the flip angle? */
	private boolean estimateFlip = false;

	/** Are we going to use a flip angle map? */
	private boolean useFlip = false;
	private String filename_flipangles = " ";
	private MCVolume alphas = null;

	/** Shall we display the short fraction after each row? */
	private boolean interactive_myelin = false;

	public void help()
	{
		super.help();

		System.out.println("\t[-f|--fit_type] <fit routine> - Parameter must be one of:");
		System.out.println("\t\tnnls");
		System.out.println("\t\tnnls_small <lower> <upper>");
		System.out.println("\t\tmonoexp");
		System.out.println("\t\tbiexp");
		System.out.println("\t\tbiexp_fixed <T2 time> <T2 time>");
		System.out.println("\t\tbiexp_sweep");
		System.out.println("\t\tl1");
		System.out.println("\t\tl1_small <lower> <upper>");

		System.out.println("");
		System.out.println("Notes:");
		System.out.println("  - Threshold specified is multiplied by standard deviation and used to"); 
		System.out.println("    threshold which pixels to map.");
		System.out.println("  - Noise estimate method can be one of:");
		System.out.println("        - 'none' - Standard deviations set to 1.0");
		System.out.println("        - 'row' - Estimate across each row");
		System.out.println("        - 'column' - Estimate across each column");
		System.out.println("        - 'image' - Estimate over each image");
		System.out.println("");

		System.out.println("Fitting Routine Notes:");
		System.out.println("  nnls - Standard minimum chi2 solution.");
		System.out.println("  nnls_small - Regularized minimum energy solution.");
		System.out.println("  monoexp - Non-linear curve fitting to find mono-exponential time and amplitude.");
		System.out.println("  biexp - Non-linear curve fitting to find bi-exponential times and amplitudes.");
		System.out.println("  biexp_fixed - Analytical solution based on notes in T2Book (Ken Whittall)");
		System.out.println("  biexp_sweep - Sweeps through a set of T2 times and finds minimum chi2 based on biexp_fixed.");
		System.out.println("  l1 - Solution from the Fortran dcl1 code."); 
		System.out.println("  l1_small - Regularized minimum energy solution form Fortran dcl1 code analogous to nnls_small."); 
	}

	/** Initializer that takes the filename.
	 *
	 */
	public void initParams( MRIFile m ) 
	{
		/*
		 *  Grab the header as we will need it 
		 *  when writing out the results.
		 */
		if( m.getFileType() == MRIFile.FILETYPE_BFF )
			bff_header = ((BFF)m).getHeader();

		if( fit_type.compareTo("nnls") == 0 )
		{
			mapper = new T2MapNNLS();
			((T2MapNNLS)mapper).setEstimateFlip( estimateFlip );
			comment += "NNLS ";
		}
		else if( fit_type.compareTo("nnls_small") == 0 )
		{
			mapper = new T2MapNNLS();
			((T2MapNNLS)mapper).setEstimateFlip( estimateFlip );
			((T2MapNNLS)mapper).setParameters(fit_params);
			comment += "NNLS (small norm " + fit_params[0]+" "+fit_params[1]+") ";
		}
		else if( fit_type.compareTo("l1_small") == 0 )
		{
			System.out.println("t2map: l1_small not implemented.");
			mapper = new T2MapL1();
			((T2MapL1)mapper).setEstimateFlip( estimateFlip );
			((T2MapL1)mapper).setParameters(fit_params);
			comment += "L1 (small norm " + fit_params[0]+" "+fit_params[1]+") ";
		}
		else if( fit_type.compareTo("monoexp") == 0 )
		{
			mapper = new T2MapMonoExp();
			comment += "mono-exponential ";
		}
		else if( fit_type.compareTo("biexp") == 0 )
		{
			mapper = new T2MapBiExp();
			comment += "bi-exponential ";
		}
		else if( fit_type.compareTo("biexp_fixed") == 0 )
		{
			mapper = new T2MapBiExpFixedT2();
			mapper.setParameters(fit_params);
			comment += "bi-exponential (fixed t2) "+fit_params[0]+" "+fit_params[1]+" ";
		}
		else if( fit_type.compareTo("biexp_sweep") == 0 )
		{
			mapper = new T2MapBiExpSweep();
			comment += "bi-exponential (sweep) ";
		}
		else if( fit_type.compareTo("l1") == 0 )
		{
			mapper = new T2MapL1();
			comment += "L1 ";
		}
		else
		{
			System.out.println("T2MapApp: Uknown fit type " + fit_type);
			System.exit(-1);
		}

		mapper.setThreshold( threshold );
		mapper.setVerbose( verbose );

		if( m.getTE() != null ) 
			mapper.setTE( m.getTE() );
		else
		{
			System.out.println("T2MapApp: Could not determine the TE's");
			System.exit(-1);
		}
		mapper.setEstimateFlip( estimateFlip );
	}
	
    public String getAppName() { return "t2map"; }
	public String getDialogName() { return "Process image"; }
	public String getOutputSuffix() { return ""; }
	public String getComment() { System.out.println("Comment is " + comment); return comment; }

	public void pre()
	{
		filterMode = READ_ONLY;
	}

	protected void moreOpts()
	{
        addOpt_noArg("normalize", 'n', "Divide by standard deviation (def: true)");
        addOpt_reqArg("output-directory", 'o', "Change output directory (def: .)");
        addOpt_reqArg("fit_type", 'f', "Set the decay curve fit type (def: nnls)");
        addOpt_reqArg("noise-estimate-method", 806, "Set the method for estimating noise (def: 'row')");
        addOpt_reqArg("flip-angle-map", 807, "Set the filename of the flip angle map (def: none)");
        addOpt_reqArg("threshold", 't', "Set the multiplier on the standard deviation (def: 3)");
        addOpt_reqArg("row-start", 800, "Starting row, 0 based (def: 0)");
        addOpt_reqArg("row-end", 801, "Ending row, 0 based (def: number rows in volume)");
        addOpt_reqArg("col-start", 802, "Starting column, 0 based (def: 0)");
        addOpt_reqArg("col-end", 803, "Ending column, 0 based (def: number column in volume)");
        addOpt_reqArg("slice-start", 804, "Starting slice, 0 based (def: 0)");
        addOpt_reqArg("slice-end", 805, "Ending slice, 0 based (def: number slices in volume)");
        addOpt_noArg("estimate-flip", 'r', "Estimate the flip angle per pixel (def: false)");
        addOpt_noArg("interactive-myelin", 'm', "Show the myelin map as we calculate it (def: false)");
	}

	protected boolean parseArg(int c)
	{
		switch(c) {
			case 'h':
				help();
				System.exit(0);
			break;
			case 'r':
				estimateFlip = true;
			break;
			case 'v':
				verbose++;
			break;
			case 807:
				useFlip = true;
				filename_flipangles = getStringArg("flip angle map filename");
			break;
			case 'n':
				normalizeFlag = false;
			break;
			case 'o':
				output_directory = getStringArg("output directory");
			break;
			case 'm':
				interactive_myelin = true;
			break;
			case 'f':
				fit_type = getStringArg("fit type");

				if( fit_type.compareTo("nnls_small") == 0 ) 
				{ 
					fit_params = new double[2];
					fit_params[0] = getDoubleArg("first fit parameter");
					fit_params[1] = getDoubleArg("second fit parameter");
				}
				else if( fit_type.compareTo("l1_small") == 0 ) 
				{ 
					fit_params = new double[2];
					fit_params[0] = getDoubleArg("first fit parameter");
					fit_params[1] = getDoubleArg("second fit parameter");
				}
				else if( fit_type.compareTo("biexp_fixed") == 0 ) 
				{ 
					fit_params = new double[2];
					fit_params[0] = getDoubleArg("first fit parameter");
					fit_params[1] = getDoubleArg("second fit parameter");
				}
			break;
			case 800:
				rowStart = getIntArg("row start");
			break;
			case 801:
				rowEnd = getIntArg("row end");
			break;
			case 802:
				colStart = getIntArg("column start");
			break;
			case 803:
				colEnd = getIntArg("column end");
			break;
			case 804:
				sliceStart = getIntArg("slice start");
			break;
			case 805:
				sliceEnd = getIntArg("slice end");
			break;
			case 806:
				ne_method = getStringArg("noise estimation method");
			break;
			case 't':
				threshold = getDoubleArg("threshold");
			break;

			default:
				help();
				System.exit(-1);
			break;
		}

		return true;
	}
		
	final public MCVolume filter( MCVolume v )
	{
		mapper.setNoiseEstimateMethod( ne_method );

		if( rowStart == -1 ) rowStart = 0;
		if( rowEnd == -1 ) rowEnd = v.getNRows();
		if( colStart == -1 ) colStart = 0;
		if( colEnd == -1 ) colEnd = v.getNCols();
		if( sliceStart == -1 ) sliceStart = 0;
		if( sliceEnd == -1 ) sliceEnd = v.getNSlices();

		if( useFlip )
		{
			BFF flips = new BFF(filename_flipangles);
			alphas = flips.getMCVolume();
			mapper.setUseFlip( alphas );
		}

		mapper.setInteractiveMyelin(interactive_myelin);
		mapper.map( v, rowStart, rowEnd, 
			  colStart, colEnd, sliceStart, sliceEnd );

		return v;
	}

	final public void post()
	{
		/*
		 *  Set the basename;
		 */
		int index = filename.indexOf(".bff");
		int lastslash = filename.lastIndexOf("/")+1;

		String basename = filename.substring(lastslash, index);
		if( output_directory.length() > 0 )
		{
			basename = output_directory + File.separator + basename;
		}
		mapper.saveData( basename , bff_header, comment );
	}
}
