package apps.t2roi;

import java.util.Vector;
import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.mri.relax.t2roi.*;
import cj.mri.relax.t2bin.*;
import cj.roi.PixFile;
import cj.appkit.FilterAppKit;
import cj.utilities.Format;

final public class T2ROIApp {
   public static void main(String argv[]) {
      T2ROIKit myApp = new T2ROIKit(); myApp.main(argv);
   }
}

final class T2ROIKit extends FilterAppKit 
{
	private Vector roifitters = new Vector();
	private Vector roifitter_names = new Vector();

	private String fit = "nnls";
	private PixFile pix = null;
	private String pix_filename = null;
	private double[] te = null;
	private int iterations = 1000;
	private double[] parameters = null;

	/** The T2 bin information. */
	private Vector t2bins = new Vector();

	/** Estimate the flip angle. */
	private boolean estimateFlip = false;

	/** Flip angle estimation method. */
	private String estimationMethod = "golden";

	/** T1. */
	private double t1 = 1000;

	private int slice = 1;

    private String comment;
	private String output_suffix = "A";

	public String getAppName() { return "t2roi"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    protected void moreOpts() { 
        addOpt_reqArg("fitter", 'f', 
			"T2 fitting method (def: nnls).");
        addOpt_reqArg("iterations", 'i', 
			"Set number of iterations for the routines (def: 1000).");
        addOpt_reqArg("measure", 'm', 
			"Routine, one of: bootstrap, mean, animal.");
        addOpt_reqArg("bin", 'b', 3,
			"Bin name, bin start (ms) and bin stop (ms).");
        addOpt_reqArg("estimate-flip-method", 800, 
			"Flip estimation method (def: minimize, must be one of 'minimize', 'spline', 'quadratic', 'golden').");
        addOpt_reqArg("t1", 801, "T1 time (def: 1000 ms).");
		addOpt_noArg("estimate-flip", 'e', "Estimate flip angle (def: false)");
    }

    protected boolean parseArg(int c)
    {
		if( t2bins == null )
		{
			t2bins = new Vector();
		}

    	switch(c) {
			case 'i':
				iterations = getIntArg("number of iterations");
				break;
			case 'e':
				estimateFlip = true;
				break;
			case 800:
				estimationMethod = getStringArg("flip angle estimation method").toLowerCase();
				break;
			case 801:
				t1 = getDoubleArg("t1");
				break;
			case 'f':
				fit = getStringArg("t2 fit routine").toLowerCase();
				
				if( fit.compareTo("nnls_small") == 0 )
				{
					parameters = new double[2];
					parameters[0] = getDoubleArg("nnls small parameter 1");
					parameters[1] = getDoubleArg("nnls small parameter 2");
				}
				break;
			case 'b':
				String name = getStringArg("bin name");
				t2bins.add( new T2Bin( getDoubleArg("bin start"), 
									   getDoubleArg("bin stop"),
									   name)); 
				break;
			case 'm':
				String measure = getStringArg("Routine").toLowerCase();

				if( measure.compareTo("bootstrap") == 0 )
				{
					roifitters.add( new T2ROIBootstrap() );
					roifitter_names.add( "Bootstrap" );
				}
				else if( measure.compareTo("mean") == 0 )
				{
					roifitters.add( new T2ROIMean() );
					roifitter_names.add( "Mean" );
				}
				else if( measure.compareTo("animal") == 0 )
				{
					roifitters.add( new T2ROIAnimal() );
					roifitter_names.add( "Animal" );
				}
				else return false;
				break;
			default:
				return false;
		}
		return true;
    }

	final public void pre()
	{
		/*
		 *  Grab the pix file which is the first
		 *  element of the otherArgs.
		 */
		pix_filename = otherArgs[0];

		if( verbose > 1 )
		{
			System.out.println("Pixel filename is " + pix_filename );
		}

		// Remove the first element.
		String[] tempArgs = (String[])otherArgs.clone();
		otherArgs = new String[ tempArgs.length - 1 ];
		for(int ii=1; ii<tempArgs.length; ii++)
		{
			otherArgs[ii-1] = tempArgs[ii];
		}

		if( verbose > 1 )
		{
			System.out.println("First bff file is " + otherArgs[0] );
		}

		if( roifitters.size() == 0 )
		{
			System.out.print("t2roi: You may want to specifiy at ");
			System.out.println("least one routine.");
			System.exit(0);
		}

		/*
		 *  Set the T2 bins.
		 */
		for(int ii=0; ii<roifitters.size(); ii++)
		{
			for(int bin=0; bin<t2bins.size(); bin++)
			{
				((T2ROI)roifitters.get(ii)).addBin(
					((T2Bin)t2bins.get(bin)).getStart(),
					((T2Bin)t2bins.get(bin)).getStop(),
					((T2Bin)t2bins.get(bin)).getName());
			}
		}

		filterMode = READ_ONLY;
	}

	final public void initParams(MRIFile b)
	{
		pix = new PixFile(b.getMCVolume().getNRows(), 
						  b.getMCVolume().getNCols() );
		pix.read( pix_filename );

		te = b.getTE();
	}

    final public void printArgs()
	{
	}

    final public MCVolume filter(MCVolume v)
	{
		System.out.println("Filename: " + filename + " PixFile: " + pix_filename);

		for(int ii=0; ii<roifitters.size(); ii++)
		{
			((T2ROI)roifitters.get(ii)).setIterations(iterations);

			if( parameters != null )
			{
				((T2ROI)roifitters.get(ii)).setParameters(parameters);
			}
			((T2ROI)roifitters.get(ii)).setVerbose(verbose);

			((T2ROI)roifitters.get(ii)).setT2FitMethod(fit);
			((T2ROI)roifitters.get(ii)).setTE(te);
			((T2ROI)roifitters.get(ii)).setT1(t1);
			((T2ROI)roifitters.get(ii)).setEstimateFlip(estimateFlip);
			((T2ROI)roifitters.get(ii)).setEstimateFlipMethod(estimationMethod);
			
			if( slice == -1 || v.getNSlices() == 1 )
				((T2ROI)roifitters.get(ii)).solve(v, pix);
			else
				((T2ROI)roifitters.get(ii)).solve(v, slice, pix);

		}

		System.out.println("");

		return v;
	}
}
